"""
Update release numbers in various places, according to a release.ini file places at the project root
"""
import configparser
import logging
from pathlib import Path
from typing import Optional, Tuple

import click
from bump_release import helpers
from bump_release.helpers import split_version

# region Constants
__version__ = VERSION = "0.9.1"
RELEASE_CONFIG: Optional[dict] = None
RELEASE_FILE: Optional[Path] = None
# endregion Constants


@click.command()
@click.option(
    "-r", "--release-file", "release_file", help="Release file path, default `./release.ini`",
)
@click.option(
    "-n", "--dry-run", "dry_run", is_flag=True, help="If set, no operation are performed on files", default=False,
)
@click.option(
    "-d", "--debug", "debug", is_flag=True, help="If set, more traces are printed for users", default=False,
)
@click.version_option(version=__version__)
@click.argument("release")
def bump_release(release: str, release_file: Optional[str] = None, dry_run: bool = False, debug: bool = False,) -> int:
    """
    Updates the files according to the release.ini file

    :param release: Version number, as "X.X.X"
    :param release_file: path to the release.ini config file
    :param dry_run: If `True`, no operation performed
    :param debug: If `True`, more traces !
    :return: 0 if no error...
    """
    # Loads the release.ini file
    global RELEASE_CONFIG, RELEASE_FILE
    if release_file is None:
        RELEASE_FILE = Path.cwd() / "release.ini"
    else:
        RELEASE_FILE = Path(release_file)

    RELEASE_CONFIG = helpers.load_release_file(release_file=RELEASE_FILE)
    return process_update(release_file=RELEASE_FILE, release=release, dry_run=dry_run, debug=debug)


def process_update(release_file: Path, release: str, dry_run: bool, debug: bool = False) -> int:
    version = split_version(release)

    # Initialize the logging
    if debug:
        logging.basicConfig(level=logging.DEBUG)
    else:
        logging.basicConfig(level=logging.INFO)

    # region Updates the main project (DJANGO_SETTINGS_MODULE file for django projects, __init__.py file...)
    try:
        new_row = update_main_file(version=version, dry_run=dry_run)
        if new_row is not None:
            logging.debug(f"process_update() `main_project`: new_row = {new_row.strip()}")
    except helpers.NothingToDoException as e:
        logging.warning(f"process_update() No release section for `main_project`: {e}")
    # endregion

    # region Updates sonar-scanner properties
    try:
        new_row = update_sonar_properties(version=version, dry_run=dry_run)
        if new_row is not None:
            logging.debug(f"process_update() `sonar`: new_row = {new_row.strip()}")
    except helpers.NothingToDoException as e:
        logging.warning(f"process_update() No release section for `sonar`: {e}")
    # endregion

    # region Updates setup.py file
    try:
        new_row = update_setup_file(version=version, dry_run=dry_run)
        if new_row is not None:
            logging.debug(f"process_update() `docs`: new_row = {new_row.strip()}")
    except helpers.NothingToDoException as e:
        logging.warning(f"process_update() No release section for `docs`: {e}")
    # endregion

    # region Updates sphinx file
    try:
        new_row = update_docs_conf(version=version, dry_run=dry_run)
        if new_row is not None:
            logging.debug(f"process_update() `docs`: new_row = {new_row.strip()}")
    except helpers.NothingToDoException as e:
        logging.warning(f"process_update() No release section for `docs`: {e}")
    # endregion

    # region Updates node packages file
    try:
        new_row = update_node_package(version=version, dry_run=dry_run)
        if new_row is not None:
            logging.debug(f"process_update() `docs`: new_row = {new_row}",)
    except helpers.NothingToDoException as e:
        logging.warning(f"process_update() No release section for `node`: {e}")
    # endregion

    # region Updates YAML file
    try:
        new_row = update_ansible_vars(version=version, dry_run=dry_run)
        if new_row is not None:
            logging.debug(f"process_update() `docs`: new_row = {new_row.strip()}")
    except helpers.NothingToDoException as e:
        logging.warning(f"process_update() No release section for `node`: {e}")
    # endregion

    # region Updates the release.ini file with the new release number
    new_row = update_release_ini(path=release_file, version=version, dry_run=dry_run)
    if new_row is not None:
        logging.warning(f"process_update() `release.ini`: new_row = {new_row.strip()}")
    # endregion

    return 0


def update_main_file(version: Tuple[str, str, str], dry_run: bool = True) -> Optional[str]:
    """
    Updates the main django settings file, or a python script with

    :param version: Release number tuple (major, minor, release)
    :param dry_run: If `True`, no operation performed
    :return: changed string
    """
    assert RELEASE_CONFIG is not None
    try:
        path = Path(RELEASE_CONFIG.get("main_project", "path"))
        pattern = RELEASE_CONFIG.get("main_project", "pattern", fallback=helpers.MAIN_PROJECT_PATTERN)  # noqa
        template = RELEASE_CONFIG.get("main_project", "template", fallback=helpers.MAIN_PROJECT_TEMPLATE)  # noqa
    except configparser.Error as e:
        raise helpers.NothingToDoException("Unable to update main project file", e)
    return helpers.update_file(path=path, pattern=pattern, template=template, version=version, dry_run=dry_run)


def update_setup_file(version: Tuple[str, str, str], dry_run: bool = False) -> Optional[str]:
    """
    Updates the setup.py file

    :param version: Release number tuple (major, minor, release)
    :param dry_run: If `True`, no operation performed
    :return: changed string
    """
    assert RELEASE_CONFIG is not None
    try:
        path = Path(RELEASE_CONFIG.get("setup", "path"))
        pattern = RELEASE_CONFIG.get("setup", "pattern", fallback=helpers.SETUP_PATTERN)  # noqa
        template = RELEASE_CONFIG.get("setup", "template", fallback=helpers.SETUP_TEMPLATE)  # noqa
    except configparser.Error as e:
        raise helpers.NothingToDoException("Unable to update setup file", e)
    return helpers.update_file(path=path, pattern=pattern, template=template, version=version, dry_run=dry_run)


def update_sonar_properties(version: Tuple[str, str, str], dry_run: bool = False) -> Optional[str]:
    """
    Updates the sonar-project.properties file with the new release number

    :param version: Release number tuple (major, minor, release)
    :param dry_run: If `True`, no operation performed
    :return: changed string
    """
    assert RELEASE_CONFIG is not None
    try:
        path = Path(RELEASE_CONFIG.get("sonar", "path"))
        pattern = RELEASE_CONFIG.get("sonar", "pattern", fallback=helpers.SONAR_PATTERN)  # noqa
        template = RELEASE_CONFIG.get("sonar", "template", fallback=helpers.SONAR_TEMPLATE)  # noqa
    except configparser.Error as e:
        raise helpers.NothingToDoException("Unable to update sonar properties file", e)
    return helpers.update_file(path=path, pattern=pattern, template=template, version=version, dry_run=dry_run)


def update_docs_conf(version: Tuple[str, str, str], dry_run: bool = False) -> Optional[str]:
    """
    Updates the Sphinx conf.py file with the new release number

    :param version: Release number tuple (major, minor, release)
    :param dry_run: If `True`, no operation performed
    :return: changed string
    """
    assert RELEASE_CONFIG is not None
    try:
        path = Path(RELEASE_CONFIG.get("docs", "path"))
        pattern_release = RELEASE_CONFIG.get("docs", "pattern_release", fallback=helpers.DOCS_RELEASE_PATTERN)  # noqa
        template_release = RELEASE_CONFIG.get("docs", "template_release", fallback=helpers.DOCS_RELEASE_FORMAT)  # noqa

        pattern_version = RELEASE_CONFIG.get("docs", "pattern_version", fallback=helpers.DOCS_VERSION_PATTERN)  # noqa
        template_version = RELEASE_CONFIG.get("docs", "re", fallback=helpers.DOCS_VERSION_FORMAT)  # noqa
    except configparser.Error as e:
        raise helpers.NothingToDoException("Unable to update docs file", e)

    update_release = helpers.update_file(
        path=path, pattern=pattern_release, template=template_release, version=version, dry_run=dry_run,
    )
    update_version = helpers.update_file(
        path=path, pattern=pattern_version, template=template_version, version=version, dry_run=dry_run,
    )
    return str(update_release) + str(update_version)


def update_node_package(version: Tuple[str, str, str], dry_run: bool = False) -> Optional[str]:
    """
    Updates the nodejs package file with the new release number

    :param version: Release number tuple (major, minor, release)
    :param dry_run: If `True`, no operation performed
    :return: changed string
    """
    assert RELEASE_CONFIG is not None
    try:
        path = Path(RELEASE_CONFIG.get("node", "path"))
        key = RELEASE_CONFIG.get("node", "key", fallback=helpers.NODE_KEY)  # noqa
    except configparser.Error as e:
        raise helpers.NothingToDoException("Unable to update node packages file", e)
    return helpers.update_node_packages(path=path, version=version, key=key, dry_run=dry_run)


def update_ansible_vars(version: Tuple[str, str, str], dry_run: bool = False) -> Optional[str]:
    """
    Updates the ansible project variables file with the new release number

    :param version: Release number tuple (major, minor, release)
    :param dry_run: If `True`, no operation performed
    :return: changed string
    """
    assert RELEASE_CONFIG is not None
    try:
        path = Path(RELEASE_CONFIG.get("ansible", "path"))
        key = RELEASE_CONFIG.get("ansible", "key", fallback=helpers.ANSIBLE_KEY)  # noqa
    except configparser.Error as e:
        raise helpers.NothingToDoException("Unable to update ansible vars file", e)
    return helpers.updates_yaml_file(path=path, version=version, key=key, dry_run=dry_run)


def update_release_ini(path: Path, version: Tuple[str, str, str], dry_run: bool = False) -> Optional[str]:
    """
    Updates the release.ini file with the new release number

    :param path: Release file path
    :param version: release number, as (<major>, <minor>, <release>)
    :param dry_run: If `True`, the operation WILL NOT be performed
    :return: Updated lines
    """
    return helpers.update_file(
        path=path,
        pattern=helpers.RELEASE_INI_PATTERN,
        template=helpers.RELEASE_INI_TEMPLATE,
        version=version,
        dry_run=dry_run,
    )



"""
__main__ for :mod:`bump_release` application

:creationdate: 17/09/2019 11:52
:moduleauthor: François GUÉRIN <fguerin@ville-tourcoing.fr>
:modulename: bump_release.__main__

"""
import sys
from pathlib import Path

import bump_release
from bump_release import helpers

__author__ = "fguerin"

if __name__ == "__main__":
    bump_release.RELEASE_FILE = Path.cwd() / "release.ini"
    bump_release.RELEASE_CONFIG = helpers.load_release_file(release_file=bump_release.RELEASE_FILE)
    result = bump_release.process_update(
        release_file=bump_release.RELEASE_FILE, release="0.4.4", dry_run=True, debug=False,
    )
    sys.exit(result)



"""
helpers for :mod:`bump_release` application

:creationdate: 16/09/2019 08:18
:moduleauthor: François GUÉRIN <fguerin@ville-tourcoing.fr>
:modulename: bump_release.helpers

"""
import configparser
import json
import logging
import os
import re
from copy import deepcopy
from pathlib import Path
from typing import Optional, Tuple, Union

from ruamel.yaml import YAML
from ruamel.yaml.compat import StringIO

__author__ = "fguerin"

RELEASE_CONFIG = None
BASE_DIR = os.getcwd()
# region Constants
# Node (JSON value update)
NODE_KEY = "release"
NODE_PACKAGE_FILE = "package.json"

# main (re search and replace)
MAIN_PROJECT_PATTERN = r"^__version__\s*=\s*VERSION\s*=\s*['\"][.\d\w]+['\"]$"
MAIN_PROJECT_TEMPLATE = '__version__ = VERSION = "{major}.{minor}.{release}"\n'

# Ansible yaml key
ANSIBLE_KEY = "git.version"

# sonar (re search and replace)
SONAR_PATTERN = r"^sonar.projectVersion=([.\d]+)$"
SONAR_TEMPLATE = "sonar.projectVersion={major}.{minor}\n"

# setup.py file
SETUP_PATTERN = r"^\s*version=['\"]([.\d\w]+)['\"],$"
SETUP_TEMPLATE = '    version="{major}.{minor}.{release}",\n'


# Sphinx (re search and replace)
DOCS_VERSION_PATTERN = r"^version\s*=\s*[\"']([.\d\w]+)[\"']$"
DOCS_RELEASE_PATTERN = r"^release\s*=\s*[\"']([.\d\w]+)[\"']$"
DOCS_VERSION_FORMAT = 'version = "{major}.{minor}"\n'
DOCS_RELEASE_FORMAT = 'release = "{major}.{minor}.{release}"\n'

RELEASE_INI_PATTERN = r"^current_release\s*=\s*['\"]?([.\d\w]+)['\"]?$"
RELEASE_INI_TEMPLATE = "current_release = {major}.{minor}.{release}\n"


# endregion Constants


def load_release_file(release_file: Union[Path, str]) -> configparser.ConfigParser:
    """
    Loads the release file and stores the config into the global :attr:`RELEASE_CONFIG`

    :param release_file: Path to the release file
    :return: Loaded config
    """
    global RELEASE_CONFIG
    RELEASE_CONFIG = configparser.ConfigParser()
    RELEASE_CONFIG.read(release_file)
    return RELEASE_CONFIG


def split_version(version: str) -> Tuple[str, str, str]:
    """
    Splits the release number into a 3-uple

    :param version: release string
    :return: <major>, <minor>, <release>
    """
    try:
        major, minor, release = version.split(".")
    except ValueError:
        logging.fatal(f'Version number "{version}" does not respect the <MAJOR>.<MINOR>.<RELEASE> format.')
        raise
    else:
        return major, minor, release


def update_file(
    path: Path, pattern: str, template: str, version: Tuple[str, str, str], dry_run: Optional[bool] = False,
) -> Optional[str]:
    """
    Performs the **real** update of the `path` files, aka. replaces the row matched
    with `pattern` with `version_format` formatted according to `release`.

    :param path: path of the file to update
    :param pattern: regexp to replace
    :param template: release format
    :param version: Release number tuple (major, minor, release)
    :param dry_run: If `True`, no operation performed
    :return: New row
    """
    version_re = re.compile(pattern)
    major, minor, release = version

    old_row, new_row = None, None
    counter = None
    with path.open(mode="r") as ifile:
        content_lines = ifile.readlines()
    new_content = deepcopy(content_lines)

    for counter, row in enumerate(content_lines):
        searched = version_re.search(row)
        if searched:
            logging.debug(f"update_file({path}) a *MATCHING* row has been found:\n{counter} {row.strip()}")
            old_row = deepcopy(row)
            new_row = template.format(major=major, minor=minor, release=release)
            break

    if old_row and new_row:
        logging.info(f"update_file({path}) old_row:\n{old_row.strip()}\nnew_row:\n{new_row.strip()}")

    if dry_run:
        logging.info(f"update_file({path}) No operation performed, dry_run = {dry_run}",)
        return new_row

    if new_row and counter is not None:
        new_content[counter] = new_row
        with path.open(mode="w") as output_file:
            output_file.writelines(new_content)
        logging.info(f"update_file({path}) File updated.")
        return new_row

    raise UpdateException(f"An error has append on updating release for file {path}")


def update_node_packages(
    path: Path, version: Tuple[str, str, str], key: str = NODE_KEY, dry_run: bool = False,
):
    """
    Updates the package.json file

    :param path: Node root directory
    :param version: Release number
    :param dry_run: If `True`, no operation performed
    :param key: json dict key (default: "release")
    :return: Nothing
    """
    try:
        with path.open(mode="r") as package_file:
            package = json.loads(package_file.read())
        new_package = deepcopy(package)
        new_package[key] = ".".join(version)
        updated = json.dumps(new_package, indent=4)
        if not dry_run:
            with path.open(mode="w") as package_file:
                package_file.write(updated)
        return updated
    except IOError as ioe:
        raise UpdateException(f"update_node_packages() Unable to perform {package_file} update: {ioe}")


class MyYAML(YAML):
    """
    Wrapper around ruamel.yaml to output directly strings
    """

    def dump(self, data, stream=None, **kw):
        inefficient = False
        if stream is None:
            inefficient = True
            stream = StringIO()
        YAML.dump(self, data, stream, **kw)
        if inefficient:
            return stream.getvalue()


def updates_yaml_file(path: Path, version: Tuple[str, str, str], key: str = ANSIBLE_KEY, dry_run: bool = False,) -> str:
    """
    Replaces the version number in a YAML file, aka. ansible vars files

    :param path: Path to the yaml file
    :param version: New version to apply, as a tuple (major, minor, release)
    :param key: key in the files, as xxx.yyy
    :param dry_run: If True, no action is performed
    :returns: new file content
    """
    splited_key = key.split(".")
    full_version = ".".join(version)
    yaml = MyYAML()
    with path.open(mode="r") as vars_file:
        document = yaml.load(vars_file)
    node = document
    for _key in splited_key:
        if _key == splited_key[-1] and not dry_run:
            node.update({_key: full_version})
        node = node.get(_key)
    logging.debug(f"updates_yml_file({vars_file}) node value = {node}")
    new_content = yaml.dump(document)
    if not dry_run:
        with path.open(mode="w") as vars_file:
            vars_file.write(new_content)
    return new_content


class UpdateException(Exception):
    """
    An error has occurred during the release updating
    """


class NothingToDoException(UpdateException):
    """
    An error has occurred during the release updating: Nothing to do...
    """



"""
setup for application

:creationdate: 13/09/2019 14:23
:moduleauthor: François GUÉRIN <fguerin@ville-tourcoing.fr>
:modulename: setup.py
"""
import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="bump_release",
    version="0.9.1",
    author="François GUÉRIN",
    author_email="fguerin@ville-tourcoing.fr",
    description="Updates various version numbers for python projects",
    long_description=long_description,
    long_description_content_type="text/markdown",
    py_modules=["bump_release"],
    install_requires=["Click", "ruamel.yaml"],
    packages=setuptools.find_packages(),
    entry_points="""
     [console_scripts]
     bump_release=bump_release:bump_release
    """,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Development Status :: 4 - Beta",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Topic :: Software Development",
    ],
    python_requires=">=3.7",
)



"""
Tests for updaters
"""
import logging
from pathlib import Path

import bump_release
import pytest
from bump_release import helpers

logger = logging.getLogger("tests")

MODULE_PATH = Path(__file__).parent


@pytest.fixture
def config():
    release_ini_path = Path.cwd() / "tests" / "fixtures" / "release.ini"
    config = helpers.load_release_file(release_ini_path)
    logger.info("config() Config file loaded")
    assert config
    return config


@pytest.fixture
def version():
    splited_version = helpers.split_version("0.0.2")
    return splited_version


def test_load_release_file(config):
    """
    Tests the loading and the values of the release.ini file

    :param config:
    :return:
    """
    assert config.has_section("main_project"), "No `main_project` section in release.ini file"
    assert config.has_section("sonar"), "No `sonar` section in release.ini file"
    assert config.has_section("docs"), "No `docs` section in release.ini file"
    assert config.has_section("ansible"), "No `ansible` section in release.ini file"

    assert config.has_option("DEFAULT", "current_release"), "No `DEFAULT.current_release` value in release.ini file"
    assert config.has_option("main_project", "path"), "No `main_project.path` value in release.ini file"
    assert config.has_option("sonar", "path"), "No `sonar.path` value in release.ini file"
    assert config.has_option("docs", "path"), "No `docs.path` value in release.ini file"
    assert config.has_option("ansible", "path"), "No `ansible.path` value in release.ini file"
    assert config.has_option("setup", "path"), "No `setup.path` value in release.ini file"


def test_version():
    splited_version = helpers.split_version("0.0.1")
    assert splited_version == ("0", "0", "1",), "Version for `0.0.1` MUST BE ('0', '0', '1')"
    splited_version = helpers.split_version("1.1.1")
    assert splited_version == ("1", "1", "1",), "Version for `1.1.1` MUST BE ('1', '1', '1')"
    splited_version = helpers.split_version("1.1.1a")
    assert splited_version == ("1", "1", "1a",), "Version for `1.1.1a` MUST BE ('1', '1', '1a')"
    splited_version = helpers.split_version("1.1.1b")
    assert splited_version == ("1", "1", "1b",), "Version for `0.0.2b` MUST BE ('1', '1', '1b')"


def test_update_main_project(config, version):
    str_path = config["main_project"].get("path")
    assert str_path is not None
    path = Path(str_path)
    new_row = helpers.update_file(
        path=path,
        pattern=helpers.MAIN_PROJECT_PATTERN,
        template=helpers.MAIN_PROJECT_TEMPLATE,
        version=version,
        dry_run=True,
    )
    assert new_row.strip() == '__version__ = VERSION = "0.0.2"', "MAIN: Versions does not match"


def test_update_sonar_properties(config, version):
    str_path = config["sonar"].get("path")
    assert str_path is not None
    path = Path(str_path)
    new_row = helpers.update_file(
        path=path, pattern=helpers.SONAR_PATTERN, template=helpers.SONAR_TEMPLATE, version=version, dry_run=True,
    )
    assert new_row.strip() == "sonar.projectVersion=0.0", "SONAR: Versions does not match"


def test_update_docs(config, version):
    str_path = config.get("docs", "path")
    assert str_path is not None
    path = Path(str_path)

    version_pattern = config.get("docs", "version_pattern", fallback=helpers.DOCS_VERSION_PATTERN)
    version_format = config.get("docs", "version_format", fallback=helpers.DOCS_VERSION_FORMAT)

    release_pattern = config.get("docs", "release_pattern", fallback=helpers.DOCS_RELEASE_PATTERN)
    release_format = config.get("docs", "release_format", fallback=helpers.DOCS_RELEASE_FORMAT)

    new_row = helpers.update_file(
        path=path, pattern=release_pattern, template=release_format, version=version, dry_run=True,
    )
    assert new_row.strip() == 'release = "0.0.2"', "DOCS: Versions does not match"

    new_row = helpers.update_file(
        path=path, pattern=version_pattern, template=version_format, version=version, dry_run=True,
    )
    assert new_row.strip() == 'version = "0.0"', "DOCS: Versions does not match"


def test_update_node_packages(config, version):
    str_path = config.get("node", "path", fallback=str(Path(__file__).parent / "fixtures" / "assets" / "package.json"),)
    assert str_path is not None
    path = Path(str_path)
    key = config.get("node", "key", fallback=helpers.NODE_KEY)
    new_content = helpers.update_node_packages(path=path, version=version, key=key)
    assert new_content, "NODE: New content cannot be empty"


def test_update_ansible(config, version):
    str_path = config.get("ansible", "path", fallback="vars.yml")
    key = config.get("ansible", "key", fallback="git.release")
    assert str_path is not None
    path = Path(str_path)
    new_content = helpers.updates_yaml_file(path=path, version=version, key=key)
    assert new_content, "ANSIBLE: New content cannot be empty"


def test_full_update_ansible(config, version):
    bump_release.RELEASE_CONFIG = config
    result = bump_release.update_ansible_vars(version=version, dry_run=True)
    assert result is not None


def test_full_docs_conf(config, version):
    bump_release.RELEASE_CONFIG = config
    result = bump_release.update_docs_conf(version=version, dry_run=True)
    assert result is not None


def test_full_main_project(config, version):
    bump_release.RELEASE_CONFIG = config
    result = bump_release.update_main_file(version=version, dry_run=True)
    assert result is not None


def test_full_node_package(config, version):
    bump_release.RELEASE_CONFIG = config
    result = bump_release.update_node_package(version=version, dry_run=True)
    assert result is not None


def test_full_sonar_properties(config, version):
    bump_release.RELEASE_CONFIG = config
    result = bump_release.update_node_package(version=version, dry_run=True)
    assert result is not None
